/*
 * Copyright (c) 2003, 2004
 *	Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 * Copyright (c) 2005 Matthieu Castet <castet.matthieu@free.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 */

/* state machine */


static int ueagle_getesi(struct ueagle_softc *, u_char *);
//static void ueagle_idma_cb(usbd_xfer_handle, usbd_private_handle, usbd_status);
static void ueagle_loadpage(void *);
static void ueagle_dispatch_cmv(struct ueagle_softc *, cmv_t *);
static usbd_status ueagle_request(struct ueagle_softc *, u_int16_t, u_int16_t,
    u_int16_t, void *);
static usbd_status ueagle_cmv(struct ueagle_softc *, u_int8_t, u_int32_t,
    u_int16_t, u_int32_t, cmv_cb_f);
static void ueagle_stat_cb(struct ueagle_softc *, int, u_int32_t);
static void ueagle_stat(void *);
static void ueagle_cntl02_cb(struct ueagle_softc *, int, u_int32_t);
static void ueagle_sync(void *);
static void ueagle_optn_cb(struct ueagle_softc *, int, u_int32_t);
static void ueagle_bootstrap(void *);
static void ueagle_endreset(void *);
static void ueagle_startreset(struct ueagle_softc *);
static int ueagle_boot(struct ueagle_softc *);
//static void ueagle_intr(usbd_xfer_handle, usbd_private_handle, usbd_status);
static int remove_timer(struct timer_list* timer);

/* Retrieve the device End System Identifier (MAC) */
#ifdef BSD
static int
ueagle_getesi(struct ueagle_softc *sc, u_char *esi)
{
	usb_string_descriptor_t us;
	u_int16_t c;
	usbd_status error;
	int i;

#if defined(__OpenBSD__)
	error = usbd_get_string_desc(sc->sc_udev, UEAGLE_ESISTRING, 0, &us);
#else
	int len;
	error = usbd_get_string_desc(sc->sc_udev, UEAGLE_ESISTRING, 0, &us,
	    &len);
#endif
	if (error != 0)
		return error;

	if (us.bLength < (6 + 1) * 2)
		return 1;

	for (i = 0; i < 6 * 2; i++) {

		if ((c = UGETW(us.bString[i])) & 0xff00)
			return 1;

		if (i & 1)
			esi[i / 2] <<= 4;
		else
			esi[i / 2] = 0;

		if (c >= '0' && c <= '9')
			esi[i / 2] |= c - '0';
		else if (c >= 'a' && c <= 'f')
			esi[i / 2] |= c - 'a' + 10;
		else if (c >= 'A' && c <= 'F')
			esi[i / 2] |= c - 'A' + 10;
		else
			return 1;
	}

	return 0;
}
#else
//GPL code
#endif
/* 
 * Set of macros to handle unaligned data in the DSP code. The DSP_GETB() macro
 * is provided only for consistency.
 */
#define DSP_GETB(p)  ((p)[0])
#define DSP_GETW(p)  ((p)[0] | (p)[1] << 8)
#define DSP_GETDW(p) ((p)[0] | (p)[1] << 8 | (p)[2] << 16 | (p)[3] << 24)

/*
 * Make sure that the DSP code provided by eaglectl is safe to use.
 */
static int
ueagle_checkdsp(u_int8_t *dsp, int len)
{
	u_int8_t  *p, *end, *pp;
	u_int8_t  pagecount, blockcount;
	u_int16_t blocksize;
	u_int32_t pageoffset;
	int i, j;

	end = dsp + len;

	/* enough space for pagecount? */
	if (dsp + 1 > end)
		return 1;

	p = dsp;
	pagecount = DSP_GETB(p);
	p += 1;

	/* enough space for page offsets? */
	if (p + 4 * pagecount > end)
		return 1;

	for (i = 0; i < pagecount; i++) {

		pageoffset = DSP_GETDW(p);
		p += 4;

		if (pageoffset == 0)
			continue;

		/* enough space for blockcount? */
		if (dsp + pageoffset + 1 > end)
			return 1;

		pp = dsp + pageoffset;
		blockcount = DSP_GETB(pp);
		pp += 1;

		for (j = 0; j < blockcount; j++) {

			/* enough space for block header? */
			if (pp + 4 > end)
				return 1;

			pp += 2; /* skip blockaddr */
			blocksize = DSP_GETW(pp);
			pp += 2;

			/* enough space for block data? */
			if (pp + blocksize > end)
				return 1;

			pp += blocksize;
		}
	}

	return 0;
}
#ifdef BSD
static void
ueagle_idma_cb(usbd_xfer_handle xfer, usbd_private_handle priv,
    usbd_status status)
{
	struct ueagle_softc *sc = priv;

	if (status != USBD_NORMAL_COMPLETION) {
		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall(sc->ep_idma);

		printf("%s: sending fw block failed: %s\n",
		    USBDEVNAME(sc->sc_dev), usbd_errstr(status));
	}

	usbd_free_xfer(xfer);
}
#else
#ifdef ASYNC_IDMA
static void
eu_idma_complete(struct urb *urb, struct pt_regs *reg) 
{
	if (urb->status) {

		printk("%s: sending DSP block failed: %d\n", 
				USBDEVNAME(sc->sc_dev), urb->status);
	}

	kfree(urb->transfer_buffer);
	usb_free_urb(urb);
}
#endif

static int eu_idma_write (struct ueagle_softc *sc, void *data, uint32_t size)
{
	int ret = 0;
	uint8_t *xfer_buf = kmalloc(size, GFP_KERNEL);


	if (!xfer_buf)
	{
		eu_err ("kmalloc in eu_idma_write for %#X bytes failed\n", size);
		ret = -ENOMEM;
		goto IDMA_exit;
	}

	memcpy(xfer_buf, data, size);
#ifdef ASYNC_IDMA
	{
	struct urb *urb;
	/*
	 * Get an URB and prepare it for submission
	 */
	urb = usb_alloc_urb(0, GFP_KERNEL);

	usb_fill_bulk_urb(urb, sc->sc_dev,
			usb_sndbulkpipe(sc->sc_dev, EP_BULK_IDMA_OUT),
			xfer_buf, size, eu_idma_complete, sc );


	ret = usb_submit_urb(urb, GFP_KERNEL);
	}
#else
	{
	int bytes_read;
	ret = usb_bulk_msg(sc->sc_dev, usb_sndbulkpipe(sc->sc_dev, EP_BULK_IDMA_OUT),
			xfer_buf, size, &bytes_read, HZ * 300);

	kfree(xfer_buf);
	if (size != bytes_read) 
		printk("size != bytes_read %d %d\n", size, bytes_read);
	}
#endif
IDMA_exit:
	return ret;
}
#endif
/*
 * The ueagle_loadpage() function must be called within a process context
 */
static void
ueagle_loadpage(void *xsc)
{
	struct ueagle_softc *sc = xsc;
	u_int16_t pageno = sc->pageno;
	u_int16_t ovl = sc->ovl;
	block_info_t bi;

	u_int8_t  *p;
	u_int8_t  pagecount, blockcount;
	u_int16_t blockaddr, blocksize;
	u_int32_t pageoffset;
	int i;

	/* reload firmware */
	if (pageno == 0 && sc->dsp) {
		release_firmware(sc->dsp_firm);
		sc->dsp = NULL;
	}
#ifdef CMVS
	if (pageno == 0 && sc->cmvs) {
		release_firmware(sc->cmv_firm);
		sc->cmvs = NULL;
		sc->max_cmvs = 0;
	}
#endif
	if (sc->dsp == NULL) {
		if (request_firmware(&sc->dsp_firm, (const char *)dsp[sc->dev], &sc->sc_dev->dev) < 0) {
			printk(KERN_ERR "Can't get firmware %s\n", dsp[sc->dev]); 
			return;
		}
		if (ueagle_checkdsp(sc->dsp_firm->data,sc->dsp_firm->size)) {
			printk(KERN_ERR "DSP firmware corrupted\n");
			return;
		}
		sc->dsp = sc->dsp_firm->data;
	}
#ifdef CMVS
	if (sc->cmvs == NULL) {
		u_int8_t *d;
		if (request_firmware(&sc->cmv_firm, (const char *)cmv[sc->dev], &sc->sc_dev->dev) < 0) {
			printk(KERN_ERR "Can't get cmv %s\n", cmv[sc->dev]); 
			return;
		}
		d = (u_int8_t*) sc->cmv_firm->data;
		if ((*d*sizeof(struct ueagle_cmvs) + 1) != sc->cmv_firm->size) {
			printk(KERN_ERR "Bad cmv %s %d %d\n", cmv[sc->dev],
					*d*sizeof(struct ueagle_cmvs)+1,
					 sc->cmv_firm->size); 
			return;
		}
		sc->max_cmvs = *d;
		printk("%u cmvs\n", sc->max_cmvs);
		sc->cmvs = (struct ueagle_cmvs*)(d + 1);
	}
#endif
	p = sc->dsp;
	pagecount = DSP_GETB(p);
	p += 1;

	if (pageno >= pagecount)
		goto bad1;

	p += 4 * pageno;
	pageoffset = DSP_GETDW(p);

	if (pageoffset == 0)
		goto bad1;

	p = sc->dsp + pageoffset;
	blockcount = DSP_GETB(p);
	p += 1;

	DPRINTF(("%s: sending %u blocks for DSP page %u\n", 
	    USBDEVNAME(sc->sc_dev), blockcount, pageno));

	USETW(bi.wHdr, UEAGLE_BIHDR);
	USETW(bi.wOvl, ovl);
	USETW(bi.wOvlOffset, ovl | 0x8000);
	for (i = 0; i < blockcount; i++) {

		blockaddr = DSP_GETW(p);
		p += 2;
		blocksize = DSP_GETW(p);
		p += 2;

		USETW(bi.wSize, blocksize);
		USETW(bi.wAddress, blockaddr);
		USETW(bi.wLast, (i == blockcount - 1) ? 1 : 0);

		/* send block info through the IDMA pipe */
		if (eu_idma_write(sc, &bi, BLOCK_INFO_SIZE)) {
			goto bad2;
		}

		/* send block data through the IDMA pipe */
		if (eu_idma_write(sc, p, blocksize)) {
			goto bad2;
		}

		p += blocksize;
	}

	return;

bad2:	printf("%s: sending DSP block %u failed\n", USBDEVNAME(sc->sc_dev), i);
	return;

bad1:	printf("%s: invalid DSP page %u requested\n", USBDEVNAME(sc->sc_dev), 
	    pageno);
}

#define ueagle_cr(sc, address, offset, callback) \
	ueagle_cmv(sc, MAKEFUNCTION(MEMACCESS, REQUESTREAD), address,	\
	    offset, 0, callback)

#define ueagle_cw(sc, address, offset, data, callback) \
	ueagle_cmv(sc, MAKEFUNCTION(MEMACCESS, REQUESTWRITE), address,	\
	    offset, data, callback)

static void
ueagle_dispatch_cmv(struct ueagle_softc *sc, cmv_t *cmv)
{
	u_int32_t data;

	if (UGETW(cmv->wPreamble) != PREAMBLE)
		goto bad1;

	if (cmv->bDirection != MODEMTOHOST)
		goto bad1;

        if (cmv->bFunction == MAKEFUNCTION(ADSLDIRECTIVE, MODEMREADY)) {
		if(!remove_timer(&sc->sc_ch))
			goto bad2;
		/* set CNTL0 to 1 and start sending configuration options */
		sc->stage = 0;
#ifdef CMVS
		if (sc->cmvs && sc->max_cmvs)
			ueagle_cw(sc, sc->cmvs[sc->stage].address,
					sc->cmvs[sc->stage].offset,
					sc->cmvs[sc->stage].data, ueagle_optn_cb);
#else
		ueagle_cw(sc, SA_CNTL, 0, 1, ueagle_optn_cb);
#endif
		return;
        }

	if (!sc->cmv_expected)
		goto bad2;

	if(!remove_timer(&sc->sc_ch))
		goto bad2;

	if (UGETW (cmv->wIndex) != sc->cmv_idx ||
	    UGETDW(cmv->dwSymbolicAddress) != sc->cmv_address ||
	    UGETW (cmv->wOffsetAddress) != sc->cmv_offset)
		goto bad2;

	if (FUNCTION_TYPE(cmv->bFunction) != MEMACCESS)
		goto bad3;

	if (FUNCTION_SUBTYPE(cmv->bFunction) != REPLYREAD &&
	    FUNCTION_SUBTYPE(cmv->bFunction) != REPLYWRITE)
		goto bad3;

	data = UGETDW(cmv->dwData);
	data = data << 16 | data >> 16;

	sc->cmv_expected = 0;
	sc->cmv_cb(sc, CMV_NORMAL_COMPLETION, data);

	return;

bad3:	sc->cmv_expected = 0;
	sc->cmv_cb(sc, CMV_ERROR, 0);
	return;

bad2:	printf("%s: unexpected cmv received\n", USBDEVNAME(sc->sc_dev));
	return;

bad1:	printf("%s: invalid cmv received\n", USBDEVNAME(sc->sc_dev));
	printf("%d %d\n", UGETW(cmv->wPreamble), cmv->bDirection);
}

#define UCDC_SEND_ENCAPSULATED_COMMAND 0x00
#ifdef BSD
static usbd_status
ueagle_request(struct ueagle_softc *sc, u_int16_t val, u_int16_t index,
    u_int16_t len, void *data)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = UCDC_SEND_ENCAPSULATED_COMMAND;
	USETW(req.wValue, val);
	USETW(req.wIndex, index);
	USETW(req.wLength, len);

	return usbd_do_request_async(sc->sc_udev, &req, data);
}
#else
static void ctrl_urb_completion(struct urb *urb, struct pt_regs *regs)
{
	struct usb_ctrlrequest *req = (struct usb_ctrlrequest *)urb->context;

	if (urb->status < 0)
		printk(KERN_DEBUG "ctrl_urb_completion() failed with %d",
				urb->status);
	kfree(req);
	kfree(urb->transfer_buffer);
	usb_free_urb(urb);
}

static int ueagle_request (struct ueagle_softc *sc, u_int16_t value,
		u_int16_t index, u_int16_t size, void *data){
	struct usb_ctrlrequest *req;
	uint8_t *xfer_buf; 
	struct urb *urb;
	int status;

	if ((urb = usb_alloc_urb(0, GFP_ATOMIC)) == NULL) {
		eu_err("Error allocating URB in ueagle_request!\n");
		return -ENOMEM;
	}

	if ((req = kmalloc(sizeof(struct usb_ctrlrequest), GFP_ATOMIC)) == NULL) {
		eu_err("Failed to allocate memory for control request\n");
		usb_free_urb(urb);
		return -ENOMEM;
	}

	//XXX malloc(0) is ok ?
	if ((xfer_buf = kmalloc(size, GFP_ATOMIC)) == NULL) {
		eu_err("Failed to allocate memory for control request\n");
		usb_free_urb(urb);
		kfree(req);
		return -ENOMEM;
	}
	memcpy(xfer_buf, data, size);
	
	req->bRequestType = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	req->bRequest = UCDC_SEND_ENCAPSULATED_COMMAND;
	req->wValue = cpu_to_le16(value);
	req->wIndex = cpu_to_le16(index);
	req->wLength = cpu_to_le16(size);

	usb_fill_control_urb (urb,
			sc->sc_dev,
			usb_sndctrlpipe(sc->sc_dev,0),
			(uint8_t *) req,
			xfer_buf,
			size,
			ctrl_urb_completion,
			req);

	if((status = usb_submit_urb(urb, GFP_ATOMIC)) < 0) {
		eu_err("Error submitting the control message: status=%d\n", status);
		kfree(req);
		kfree(xfer_buf);
		usb_free_urb(urb);
	}
	return status;
}

#define usb_callout(a,b,c,d) usb_callout1(&a,b,(void *)c,d)
#define MS_TO_TICKS(x) MSEC_TO_JIFFIES(x)
static int remove_timer(struct timer_list* timer) {
	/*If the timer is currently running, delete it*/
	if (timer_pending(timer)) {
		del_timer(timer);
		return 1;
	}
	else
		return 0;
}
static void usb_callout1(struct timer_list* timer, int reqdelay, void (*f)(void *p), struct ueagle_softc *sc) 
{
	remove_timer(timer);

	timer->function = (void *) f;
	timer->data = (unsigned long) sc;
	timer->expires = reqdelay + jiffies;

	add_timer(timer);
}
#endif

static usbd_status
ueagle_cmv(struct ueagle_softc *sc, u_int8_t function, u_int32_t address,
    u_int16_t offset, u_int32_t data, cmv_cb_f cb)
{
	cmv_t cmv;

	USETW (cmv.wPreamble, PREAMBLE);
	cmv.bDirection = HOSTTOMODEM;
	cmv.bFunction = function;
	sc->cmv_idx++;
	USETW (cmv.wIndex, sc->cmv_idx);
	USETDW(cmv.dwSymbolicAddress, address);
	USETW (cmv.wOffsetAddress, offset);
	data = data >> 16 | data << 16;
	USETDW(cmv.dwData, data);

	sc->cmv_expected = 1;
	sc->cmv_address = address;
	sc->cmv_offset = offset;
	sc->cmv_cb = cb;

	if (timer_pending(&sc->sc_ch)) {
		eu_err ("TxPending = TRUE, cannot send new message!\n");
	}
	usb_callout(sc->sc_ch, MS_TO_TICKS(3000), ueagle_startreset, sc);

	return ueagle_request(sc, UEAGLE_SETBLOCK, UEAGLE_MPTXSTART, CMV_SIZE,
	    &cmv);
}

static void
ueagle_stat_cb(struct ueagle_softc *sc, int status, u_int32_t data)
{
	u_int32_t old_state;
	if (status != CMV_NORMAL_COMPLETION) {
		printf("%s: failed to receive CMV ack: %d\n",
		    USBDEVNAME(sc->sc_dev), status);

		ueagle_startreset(sc);
		return;
	}

	switch (sc->stage) {
	case 0: /* STAT0 ack */
		old_state = sc->stats.phy.state;
		sc->stats.phy.state = data;

		switch ((data >> 8) & 0xf) {
		case 0: /* not yet synchronized */
			DPRINTFN(3, ("%s: modem not yet synchronized\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_sync(sc); /* retry */
			return;

		case 1: /* initialization */
			DPRINTFN(3, ("%s: modem initializing\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_sync(sc); /* retry */
			return;

		case 2: /* operational */
			DPRINTFN(4, ("%s: modem operational\n",
			    USBDEVNAME(sc->sc_dev)));
			/* sometimes the modem set LOOPBACKON when rebboting */
			//if ((((old_state >> 8) & 0xf) != 2)/* && sc->running*/) {
        			ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKOFF, 0, NULL);
			//	eu_report("loopback off\n");
			//}
			UPDATE_ATM_STAT(signal, ATM_PHY_SIG_FOUND);
			ueagle_cr(sc, SA_DIAG, 2, ueagle_stat_cb);
			break;

		default: /* fail ... */
			DPRINTFN(3, ("%s: modem synchronization failed :%d\n",
			    USBDEVNAME(sc->sc_dev), (data >> 8) & 0xf));
			ueagle_startreset(sc);
			return;
		}
		break;

	case 1: /* DIAG1 ack */
		sc->stats.phy.flags = data;
		sc->stats.phy.mflags |= data;

		if (data & 0x10) {
			DPRINTF(("%s: delineation LOSS\n",
			    USBDEVNAME(sc->sc_dev)));
			ueagle_startreset(sc);
			return;
		}

		ueagle_cr(sc, SA_RATE, 0, ueagle_stat_cb);
		break;

	case 2: /* RATE0 ack */
#ifndef TIMEOUT
		/* in could mode the modem have problem with high rate
		 * changing internal timing could improve things, but the 
		 * value is misterious
		 */
		if (!iso[sc->dev] && sc->stats.phy.dsrate != (data >> 16) * 32) {
			int timeout;
#if 0
			if (data <= 0x8000)
				timeout = 0xb;
			else
				timeout = 0xf;
#else
			if (data <= 0x2000)
				timeout = 0;
			else
				timeout = 1;
#endif
			ueagle_request(sc, UEAGLE_SETTIMEOUT, timeout, 0, NULL);
			eu_report("Setting new timeout1 %d\n", timeout);
		}
#endif
		sc->stats.phy.dsrate = (data >> 16) * 32;
		sc->stats.phy.usrate = (data & 0xffff) * 32;
			
		UPDATE_ATM_STAT(link_rate, sc->stats.phy.dsrate * 1000 / 424);
		ueagle_cr(sc, SA_DIAG, 3, ueagle_stat_cb);
		break;

	case 3:	/* DIAG3 ack */
		sc->stats.phy.cells_crc_errors = data;
		ueagle_cr(sc, SA_DIAG, 22, ueagle_stat_cb);
		break;

	case 4: /* DIAG22 ack */
		sc->stats.phy.dserror = data;
		ueagle_cr(sc, SA_DIAG, 23, ueagle_stat_cb);
		break;

	case 5: /* DIAG23 ack */
		sc->stats.phy.attenuation = (data & 0xff) / 2;
		ueagle_cr(sc, SA_DIAG, 25, ueagle_stat_cb);
		break;

	case 6: /* DIAG25 ack */
		sc->stats.phy.dsmargin = data;
		ueagle_cr(sc, SA_DIAG, 46, ueagle_stat_cb);
		break;

	case 7: /* DIAG46 ack */
		sc->stats.phy.userror = data;
		ueagle_cr(sc, SA_DIAG, 49, ueagle_stat_cb);
		break;

	case 8: /* DIAG49 ack */
		sc->stats.phy.usmargin = data;
		ueagle_cr(sc, SA_DIAG, 51, ueagle_stat_cb);
		break;

	case 9: /* DIAG51 ack */
		sc->stats.phy.rxflow = data;
		ueagle_cr(sc, SA_DIAG, 52, ueagle_stat_cb);
		break;

	case 10: /* DIAG52 ack */
		sc->stats.phy.txflow = data;
		ueagle_cr(sc, SA_DIAG, 54, ueagle_stat_cb);
		break;
		/* DIAG53 ack */
		/* near_end_Current_Corrected_Blocks */

	case 11: /* DIAG54 ack */
		sc->stats.phy.dsunc = data;
		ueagle_cr(sc, SA_DIAG, 58, ueagle_stat_cb);
		break;

	case 12: /* DIAG58 ack */
		sc->stats.phy.usunc = data;
		ueagle_cr(sc, SA_INFO, 8, ueagle_stat_cb);
		break;

	case 13: /* INFO8 ack */
		sc->stats.phy.vidco = data;
		ueagle_cr(sc, SA_INFO, 14, ueagle_stat_cb);
		break;

	case 14: /* INFO14 ack */
		sc->stats.phy.vidcpe = data;
		sc->stage = 0;
		usb_callout(sc->sc_ch, MS_TO_TICKS(1000), ueagle_stat, sc);

		/* wake up processes waiting for synchronization */
		wakeup(sc);
		return;
	}

	sc->stage++;
}

/* Start requesting statistics to the modem */
static void
ueagle_stat(void *priv)
{
	struct ueagle_softc *sc = priv;

	DPRINTFN(4, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	ueagle_cr(sc, SA_STAT, 0, ueagle_stat_cb);
}

static void
ueagle_cntl02_cb(struct ueagle_softc *sc, int status, u_int32_t data)
{
	if (status != CMV_NORMAL_COMPLETION) {
		printf("%s: failed to receive CMV ack: %d\n",
		    USBDEVNAME(sc->sc_dev), status);

		ueagle_startreset(sc);
		return;
	}

	sc->stage = 0;

	/* can't sleep since we are called from an interrupt context */
        usb_callout(sc->sc_ch, MS_TO_TICKS(10000), ueagle_stat, sc);
}

static void
ueagle_sync(void *priv)
{
	struct ueagle_softc *sc = priv;

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	UPDATE_ATM_STAT(signal, ATM_PHY_SIG_LOST);
	//ueagle_cw(sc, SA_CNTL, 0, 2, ueagle_cntl02_cb);
	ueagle_cntl02_cb(sc, CMV_NORMAL_COMPLETION, 0);
}

static void
ueagle_optn_cb(struct ueagle_softc *sc, int status, u_int32_t data)
{
	if (status != CMV_NORMAL_COMPLETION) {
		printf("%s: failed to receive CMV ack: %d\n",
		    USBDEVNAME(sc->sc_dev), status);

		ueagle_startreset(sc);
		return;
	}

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

#ifndef CMVS
	switch (sc->stage) {
	case 0: /* CNTL0 ack */
		sc->stage = 1;
		ueagle_cw(sc, SA_OPTN, 0, UEAGLE_OPTN0, ueagle_optn_cb);
		break;

        case 1: /* OPTN0 ack */
		sc->stage = 2;
		ueagle_cw(sc, SA_OPTN, 2, UEAGLE_OPTN2, ueagle_optn_cb);
		break;

	case 2:
#ifdef UEAGLE_OPTN3
		sc->stage = 3;
		ueagle_cw(sc, SA_OPTN, 3, UEAGLE_OPTN3, ueagle_optn_cb);
		break;

#endif
	case 3:
#ifdef UEAGLE_OPTN4
		sc->stage = 4;
		ueagle_cw(sc, SA_OPTN, 4, UEAGLE_OPTN4, ueagle_optn_cb);
		break;

#endif
	case 4:
#ifdef UEAGLE_OPTN5
		sc->stage = 5;
		ueagle_cw(sc, SA_OPTN, 5, UEAGLE_OPTN5, ueagle_optn_cb);
		break;

#endif
	case 5:
#ifdef UEAGLE_OPTN6
		sc->stage = 6;
		ueagle_cw(sc, SA_OPTN, 6, UEAGLE_OPTN6, ueagle_optn_cb);
		break;

#endif
        case 6:
		sc->stage = 7;
		ueagle_cw(sc, SA_OPTN, 7, UEAGLE_OPTN7, ueagle_optn_cb);
		break;

	case 7:
#ifdef UEAGLE_OPTN15
		sc->stage = 8;
		ueagle_cw(sc, SA_OPTN, 15, UEAGLE_OPTN15, ueagle_optn_cb);
		break;

#endif
        case 8:
		ueagle_sync(sc);
		return;
        }
#else
	sc->stage++;
	if (sc->stage >= sc->max_cmvs) {
		/* CMVS already send cntl02*/
		ueagle_cntl02_cb(sc, CMV_NORMAL_COMPLETION, 0);
		return;
	}
	ueagle_cw(sc, sc->cmvs[sc->stage].address, sc->cmvs[sc->stage].offset, sc->cmvs[sc->stage].data, ueagle_optn_cb);
#endif

}

static void
ueagle_bootstrap(void *priv)
{
	struct ueagle_softc *sc = priv;

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	/*
	 * we start a timeout in case the modem fails to send a
	 * modem ready CMV.
	 */
	usb_callout(sc->sc_ch, MS_TO_TICKS(8000), ueagle_startreset, sc);
	sc->booting = 0;

	sc->pageno = 0;
	sc->ovl = 0;
	schedule_work (&sc->sc_task);
}

/*-
 * 1) leave reset mode
 * 2) clear tx and rx mailboxes
 * 3) wait 1s before boot straping the firmware
 */
static void
ueagle_endreset(void *priv)
{
	struct ueagle_softc *sc = priv;
	u_int16_t zero = 0; /* ;-) */

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_ENDRESET, 0, NULL);
        ueagle_request(sc, UEAGLE_SET2183DATA, UEAGLE_MPTXMAILBOX, 2, &zero);
        ueagle_request(sc, UEAGLE_SET2183DATA, UEAGLE_MPRXMAILBOX, 2, &zero);
        ueagle_request(sc, UEAGLE_SET2183DATA, UEAGLE_SWAPMAILBOX, 2, &zero);


	/* can't sleep since we can be called from an interrupt context */
        usb_callout(sc->sc_ch, MS_TO_TICKS(1000), ueagle_bootstrap, sc);
}

/*-
 * 1) tell the modem that we want to boot in IDMA mode
 * 2) enter reset mode
 * 3) wait 200ms before leaving reset mode
 */
static void
ueagle_startreset(struct ueagle_softc *sc)
{
	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	UPDATE_ATM_STAT(signal, ATM_PHY_SIG_LOST);
	sc->stats.phy.state = 0;
	sc->booting = 1;

	ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKON, 0, NULL);
        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_BOOTIDMA, 0, NULL);
        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_STARTRESET, 0, NULL);

	/* can't sleep since we can be called from an interrupt context */
        usb_callout(sc->sc_ch, MS_TO_TICKS(100), ueagle_endreset, sc);
}

#ifdef BSD
/* open the IDMA and interrupt pipes */
static void
ueagle_boot(struct ueagle_softc *sc)
{
	usbd_interface_handle iface;
	char *errmsg;

	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	if (usbd_device2interface_handle(sc->sc_udev, UEAGLE_US_IFACE_NO,
	    &iface) != 0) {
		errmsg = "getting idma interface handle failed";
		goto bad;
	}

	if (usbd_open_pipe(iface, UEAGLE_IDMA_PIPE, USBD_EXCLUSIVE_USE,
	    &sc->ep_idma) != 0) {
		errmsg = "opening idma pipe failed";
		goto bad;
	}

	if (usbd_device2interface_handle(sc->sc_udev, UEAGLE_INTR_IFACE_NO,
	    &iface) != 0) {
		errmsg = "getting interrupt interface handle failed";
		goto bad;
	}

	if ((sc->ibuf = malloc(INTR_PKT_SIZE, M_USBDEV, M_NOWAIT)) == NULL) {
		errmsg = "allocating interrupt buffer failed";
		goto bad;
	}

	if (usbd_open_pipe_intr(iface, UEAGLE_INTR_PIPE, USBD_SHORT_XFER_OK,
	    &sc->ep_intr, sc, sc->ibuf, INTR_PKT_SIZE, ueagle_intr,
	    UEAGLE_INTR_INTERVAL) != 0) {
		errmsg = "opening interrupt pipe failed";
		goto bad;
	}
        ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_BOOTIDMA, 0, NULL);
	ueagle_request(sc, UEAGLE_SETMODE, UEAGLE_LOOPBACKON, 0, NULL);
	msleep(100);

	ueagle_startreset(sc);

	return;

bad:	printf("%s: %s\n", USBDEVNAME(sc->sc_dev), errmsg);
	ueagle_stop(sc);
}

static void
ueagle_intr(usbd_xfer_handle xfer, usbd_private_handle priv, usbd_status status)
{
	struct ueagle_softc *sc = priv;
	intr_pkt_t *intr;

	if (status != USBD_NORMAL_COMPLETION) {
		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall_async(sc->ep_intr);

		return;
	}

	intr = sc->ibuf;

	switch (UGETW(intr->wInterrupt)) {
	case INT_LOADSWAPPAGE:
#if defined(__NetBSD__) || defined(__OpenBSD__)
		sc->pageno = intr->bSwapPageNo;
		sc->ovl = intr->bOvl >> 4 | intr->bOvl << 4;
		usb_add_task(sc->sc_udev, &sc->sc_task);
#elif defined(__FreeBSD__)
		ueagle_loadpage(sc, intr->bSwapPageNo,
		    intr->bOvl >> 4 | intr->bOvl << 4);
#endif
		break;

	case INT_INCOMINGCMV:
		ueagle_dispatch_cmv(sc, &intr->u.s2.cmv);
		break;

	default:
		printf("%s: unknown intr %u\n", USBDEVNAME(sc->sc_dev),
		    UGETW(intr->wInterrupt));
	}
}
#else
static void ueagle_intr (struct urb *urb, struct pt_regs *regs)
{
	struct ueagle_softc *sc = (struct ueagle_softc *)urb->context;
	intr_pkt_t *intr;

	if (urb->status < 0) {
		printk(KERN_DEBUG "ueagle_intr() failed with %d\n",
				urb->status);
		return;
	}

	intr = sc->ibuf;

	/* device-to-host interrupt */
	if (intr->bType != 0x08 || sc->booting) {
		printk("wrong int\n");
		// rebooting ?
		goto resubmit;
	}


	switch (UGETW(intr->wInterrupt)) {
	case INT_LOADSWAPPAGE:
		sc->pageno = intr->bSwapPageNo;
		sc->ovl = intr->bOvl >> 4 | intr->bOvl << 4;
		schedule_work (&sc->sc_task);
		break;

	case INT_INCOMINGCMV:
		ueagle_dispatch_cmv(sc, &intr->u.s2.cmv);
		break;

	default:
		printf("%s: unknown intr %u\n", USBDEVNAME(sc->sc_dev), 
		    UGETW(intr->wInterrupt));
	}
resubmit:
	usb_submit_urb(sc->urb_int, GFP_ATOMIC);
}

static int
ueagle_boot(struct ueagle_softc *sc)
{
	DPRINTFN(3, ("%s: %s\n", USBDEVNAME(sc->sc_dev), __func__));

	if ((sc->ibuf = kmalloc(INTR_PKT_SIZE, GFP_ATOMIC)) == NULL) {
		return -ENOMEM;
	}

	if (!(sc->urb_int = usb_alloc_urb(0, GFP_KERNEL))) {
		eu_err("cannot allocate interrupt URB\n");
		goto err;
	}

	usb_fill_int_urb(sc->urb_int, sc->sc_dev,
			usb_rcvintpipe(sc->sc_dev, EP_INT_IN),
			sc->ibuf, INTR_PKT_SIZE, ueagle_intr, sc,
			sc->sc_dev->actconfig->interface[0]->altsetting[0].endpoint[0].desc.bInterval);
	if (usb_submit_urb(sc->urb_int, GFP_KERNEL) < 0) {
		eu_err("Failed to submit interrupt URB\n");
		goto err1;
	}

	ueagle_startreset(sc);

	return 0;

err1:	
	kfree(sc->ibuf);
err:	
	usb_free_urb(sc->urb_int);
	return -ENOMEM;
}

static void ueagle_stop(struct ueagle_softc *sc)
{
	flush_scheduled_work();
	/* stop any pending boot process */
	del_timer(&sc->sc_ch);


	usb_kill_urb(sc->urb_int);
	usb_free_urb(sc->urb_int);
    	kfree(sc->ibuf);

	/* reset statistics */
	memset(&sc->stats, 0, sizeof (struct ueagle_stats));
#ifdef CMVS
	if (sc->cmv_firm)
		release_firmware(sc->cmv_firm);
#endif
	if (sc->dsp_firm)
		release_firmware(sc->dsp_firm);
}



static int init(struct ueagle_softc *sc) {

	sc->ibuf = NULL;
	//sc->gone = 0;
	sc->dsp = NULL;
	sc->cmv_idx = 0;
	sc->cmv_expected = 0;
	memset(&sc->stats, 0, sizeof (struct ueagle_stats));
	INIT_WORK(&sc->sc_task, ueagle_loadpage, sc);
	init_timer(&sc->sc_ch);
	init_waitqueue_head(&sc->sync_q);
#ifdef CMVS
	sc->max_cmvs = 0;
	sc->cmvs = NULL;
#endif
	sc->booting = 1;

	return 0;
}
#endif

