/*-
 * Copyright (c) 2003, 2004
 *	Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 * 
 * Copyright (c) 2005 Matthieu Castet <castet.matthieu@free.fr>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define CMVS
#include "usbatm.h"

#if defined(__NetBSD__) || defined(__OpenBSD__)
#define MS_TO_TICKS(ms) mstohz(ms)
#elif defined(__FreeBSD__)
#define MS_TO_TICKS(ms) ((ms) * hz / 1000)
#endif

#if defined(__FreeBSD__) && __FreeBSD_version < 500000
typedef struct callout_handle	usb_callout_t;
#define usb_callout_init(h)	callout_handle_init(&(h))
#define usb_callout(h, t, f ,d)	usb_timeout((f), (d), (t), (h))
#define usb_uncallout(h, f, d)	usb_untimeout((f), (d), (h))
#endif

#if defined(__FreeBSD__) && __FreeBSD_version >= 501000
#define GET_IFP(sc)	(&(sc)->ifatm.ifnet)
#else
#define GET_IFP(sc)	(&(sc)->ifnet)
#endif

#if defined(__FreeBSD__) && __FreeBSD_version < 500000
#define BPF_MTAP(ifp, m) do {						\
	if ((ifp)->if_bpf)						\
		bpf_mtap((ifp), (m));					\
} while (/* CONSTCOND */0)
#elif defined(__NetBSD__) || defined(__OpenBSD__)
#define BPF_MTAP(ifp, m) do {						\
	if ((ifp)->if_bpf)						\
		bpf_mtap((ifp)->if_bpf, (m));				\
} while (/* CONSTCOND */0)
#endif

#define UEAGLE_NISOREQS		6
#define UEAGLE_NISOFRMS		4

#define UEAGLE_TX_LIST_CNT	1

#define UEAGLE_MAXDSPSIZE	(1024 * 1024)

#define UEAGLE_IDMA_TIMEOUT	10000
#define UEAGLE_TX_TIMEOUT	10000

#define UEAGLE_IFMTU		1500

struct ueagle_softc;

#define CMV_NORMAL_COMPLETION	0
#define CMV_TIMEOUT		1
#define CMV_ERROR		2
typedef void (*cmv_cb_f)(struct ueagle_softc *, int, u_int32_t);

#ifdef CMVS
struct ueagle_cmvs {
	u_int32_t address;
	u_int16_t offset;
	u_int32_t data;
} __attribute__ ((packed));
#endif

struct ueagle_softc {
	struct usb_device*	sc_dev;
	struct usbatm_data 	*usbatm;
	int			dev;
	unsigned int		driver_info;

	int			running;
	int			booting;

	u_int8_t		stage;

	struct timer_list	sc_ch;

	wait_queue_head_t	sync_q;

	struct work_struct      sc_task;
	u_int16_t		pageno;
	u_int16_t		ovl;

	u_int8_t		*dsp;
	const struct firmware 	*dsp_firm;
	void			*ibuf;
	struct urb*		urb_int;
	u_int			isize;

	u_int16_t		cmv_idx;
	u_int16_t		cmv_expected;
	u_int32_t 		cmv_address;
	u_int16_t		cmv_offset;
	cmv_cb_f		cmv_cb;
#ifdef CMVS
	u_int8_t		max_cmvs;
	struct ueagle_cmvs	*cmvs;
	const struct firmware   *cmv_firm;
#endif

	/* keep in sync with eaglectl */
	struct ueagle_stats {
		struct {
			u_int32_t	state;
			u_int32_t	flags;
			u_int32_t	mflags;
			u_int32_t	vidcpe;
			u_int32_t	vidco;
			u_int32_t	dsrate;
			u_int32_t	usrate;
			u_int32_t	dserror;
			u_int32_t	userror;
			u_int32_t	dsunc;
			u_int32_t	usunc;
			u_int32_t	txflow;
			u_int32_t	rxflow;
			u_int32_t	attenuation;
			u_int32_t	dsmargin;
			u_int32_t	usmargin;
			u_int32_t	cells_crc_errors;
		} phy;
	} stats;
};

/* constants and structures for ioctl stuff (keep in sync with eaglectl) */

struct eagle_dsp {
	caddr_t code;
	int len;
};
